"""
Contains all functions that access any functionCall-object
"""

import time

from sqlalchemy import func

from flask_monitoringdashboard.database import Request


def add_request(db_session, duration, endpoint_id, ip):
    """ Adds a request to the database. Returns the id."""
    request = Request(endpoint_id=endpoint_id, duration=duration, ip=ip)
    db_session.add(request)
    db_session.flush()
    return request.id


def get_data_between(db_session, time_from, time_to=None):
    """
        Returns all data in the Request table, for the export data option.
        This function returns all data after the time_from date.
    """
    query = db_session.query(Request).filter(Request.time_requested > time_from)
    if time_to:
        query = query.filter(Request.time_requested <= time_to)
    return query.all()


def get_data(db_session):
    """
    Equivalent function to get_data_from, but returns all data.
    :return: all data from the database in the Endpoint-table.
    """
    return db_session.query(Request).all()


def get_date_of_first_request(db_session):
    """ return the date (as unix timestamp) of the first request """
    result = db_session.query(Request.time_requested).order_by(Request.time_requested).first()
    if result:
        return int(time.mktime(result[0].timetuple()))
    return -1


def get_avg_execution_time(db_session, endpoint_id):
    """ Return the average execution time of an endpoint """
    result = db_session.query(func.avg(Request.duration).label('average')). \
        filter(Request.endpoint_id == endpoint_id).one()
    if result[0]:
        return result[0]
    return 0  # default value



"""
    This file can be executed for developing purposes.
    It is not used when the flask_monitoring_dashboard is attached to an existing flask application.
"""

from flask import Flask


def create_app():
    import flask_monitoringdashboard as dashboard
    import time

    app = Flask(__name__)

    dashboard.config.outlier_detection_constant = 0
    dashboard.config.database_name = 'sqlite:///flask_monitoring_dashboard_v10.db'
    dashboard.bind(app)

    def f(duration=1):
        time.sleep(duration)

    def g():
        f()

    @app.route('/endpoint')
    def endpoint():
        if random.randint(0, 1) == 0:
            g()
        else:
            f()
        return 'Ok'

    return app


if __name__ == '__main__':
    create_app().run(debug=True)



"""
    Helper functions for parsing the arguments from the config file
"""
import ast
import os


def parse_version(parser, header, version):
    """
    Parse the version given in the config-file.
    If both GIT and VERSION are used, the GIT argument is used.
    :param parser: the parser to be used for parsing
    :param header: name of the header in the configuration file
    :param version: the default version
    """
    version = parse_string(parser, header, 'APP_VERSION', version)
    if parser.has_option(header, 'GIT'):
        git = parser.get(header, 'GIT')
        try:
            # current hash can be found in the link in HEAD-file in git-folder
            # The file is specified by: 'ref: <location>'
            git_file = (open(os.path.join(git, 'HEAD')).read().rsplit(': ', 1)[1]).rstrip()
            # read the git-version
            version = open(git + '/' + git_file).read()
            # cut version to at most 6 chars
            return version[:6]
        except IOError:
            print("Error reading one of the files to retrieve the current git-version.")
            raise
    return version


def parse_string(parser, header, arg_name, arg_value):
    """
    Parse an argument from the given parser. If the argument is not specified, return the default value
    :param parser: the parser to be used for parsing
    :param header: name of the header in the configuration file
    :param arg_name: name in the configuration file
    :param arg_value: default value, the the value is not found
    """
    if parser.has_option(header, arg_name):
        return parser.get(header, arg_name)
    return arg_value


def parse_bool(parser, header, arg_name, arg_value):
    """
    Parse an argument from the given parser. If the argument is not specified, return the default value
    :param parser: the parser to be used for parsing
    :param header: name of the header in the configuration file
    :param arg_name: name in the configuration file
    :param arg_value: default value, the the value is not found
    """
    if parser.has_option(header, arg_name):
        return parser.get(header, arg_name) == 'True'
    return arg_value


def parse_literal(parser, header, arg_name, arg_value):
    """
    Parse an argument from the given parser. If the argument is not specified, return the default value
    :param parser: the parser to be used for parsing
    :param header: name of the header in the configuration file
    :param arg_name: name in the configuration file
    :param arg_value: default value, the the value is not found
    """
    if parser.has_option(header, arg_name):
        return ast.literal_eval(parser.get(header, arg_name))
    return arg_value



"""
    This file contains all unit tests that count a number of results in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/count.py')
    See info_box.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count import count_versions_endpoint
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, ENDPOINT_ID


class TestCount(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_count_versions(self):
        with session_scope() as db_session:
            self.assertEqual(count_versions_endpoint(db_session, ENDPOINT_ID), 1)



"""
    Core files for the Flask Monitoring Dashboard
    - auth.py handles authentication
    - forms.py are used for generating WTF_Forms
    - measurements.py contains a number of wrappers
    - outlier.py contains outlier information
"""


from sqlalchemy import desc, asc
from sqlalchemy.orm import joinedload

from flask_monitoringdashboard.database import Outlier, Request


def add_outlier(db_session, request_id, cpu_percent, memory, stacktrace, request):
    """ Collects information (request-parameters, memory, stacktrace) about the request and adds it in the database."""
    headers, environ, url = request
    outlier = Outlier(request_id=request_id, request_header=headers,
                      request_environment=environ,
                      request_url=url, cpu_percent=cpu_percent,
                      memory=memory, stacktrace=stacktrace)
    db_session.add(outlier)


def get_outliers_sorted(db_session, endpoint_id, offset, per_page):
    """
    :param db_session: the session containing the query
    :param endpoint_id: endpoint_id for filtering the requests
    :param offset: number of items to skip
    :param per_page: number of items to return
    :return: a list of all outliers of a specific endpoint. The list is sorted based on request time.
    """
    result = db_session.query(Outlier).\
        join(Outlier.request). \
        options(joinedload(Outlier.request).joinedload(Request.endpoint)). \
        filter(Request.endpoint_id == endpoint_id). \
        order_by(desc(Request.time_requested)). \
        offset(offset).limit(per_page).all()
    db_session.expunge_all()
    return result


def get_outliers_cpus(db_session, endpoint_id):
    """
    :param db_session: the session containing the query
    :param endpoint_id: endpoint_id for filtering the requests
    :return: a list of all cpu percentages for outliers of a specific endpoint
    """
    outliers = db_session.query(Outlier.cpu_percent).\
        join(Outlier.request).\
        filter(Request.endpoint_id == endpoint_id).all()
    return [outlier[0] for outlier in outliers]




import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count import count_requests
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, get_test_app

NAME = 'test_endpoint'


class TestMeasurement(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    @staticmethod
    def func():
        """ function for testing the performance """
        return 'Hello World!'

    def test_get_group_by(self):
        """
            Test whether the group_by returns the right result
        """
        from flask_monitoringdashboard import config
        from flask_monitoringdashboard.core.group_by import get_group_by
        config.group_by = lambda: 3
        self.assertEqual(get_group_by(), '3')

        config.group_by = (lambda: 'User', lambda: 3.0)
        self.assertEqual(get_group_by(), '(User,3.0)')

    # def test_track_performance(self):
    #     """
    #         Test whether the track_performance works
    #     """
    #     from flask_monitoringdashboard.core.measurement import track_performance
    #
    #     with self.app.test_request_context(environ_base={'REMOTE_ADDR': '127.0.0.1'}):
    #         self.func = track_performance(self.func, NAME)
    #         self.func()
    #         with session_scope() as db_session:
    #             self.assertEqual(count_requests(db_session, NAME), 1)

    # def test_add_decorator(self):
    #     """
    #         Test whether the add_decorator works
    #     """
    #     from flask_monitoringdashboard.core.measurement import add_decorator
    #
    #     with self.app.test_request_context(environ_base={'REMOTE_ADDR': '127.0.0.1'}):
    #         self.func = add_decorator(NAME, monitor_level=1)
    #         self.func()
    #         with session_scope() as db_session:
    #             self.assertEqual(count_requests(db_session, NAME), 1)

    # def test_last_accessed(self):
    #     """
    #         Test whether the last_accessed is stored in the database
    #     """
    #     from flask_monitoringdashboard.core.measurement import track_last_accessed
    #     from flask_monitoringdashboard.database.count_group import get_value
    #
    #     with session_scope() as db_session:
    #         get_monitor_rule(db_session, NAME)
    #         func = track_last_accessed(self.func, NAME)
    #
    #     with session_scope() as db_session:
    #         self.assertIsNone(get_value(get_last_accessed_times(db_session), NAME, default=None))
    #         func()
    #         self.assertIsNotNone(get_value(get_last_accessed_times(db_session), NAME, default=None))

    # def test_get_average(self):
    #     """
    #         Test get_average
    #     """
    #     from flask_monitoringdashboard.core.measurement import track_performance, get_average, MIN_NUM_REQUESTS
    #     self.assertIsNone(get_average(NAME))
    #
    #     with self.app.test_request_context(environ_base={'REMOTE_ADDR': '127.0.0.1'}):
    #         self.func = track_performance(self.func, NAME)
    #         for i in range(MIN_NUM_REQUESTS):
    #             self.func()
    #
    #     self.assertIsNotNone(get_average(NAME))



"""
    Create a StackInfo-object and it will spawn a different thread.
    That thread will sleep for 'average' ms (reformatted to seconds).
    Then it will log the stacktrace of all active threads.
    Moreover, it logs cpu- and memory-info.
"""

import time
import traceback
from threading import Thread
import sys
import psutil


class StackInfo(object):
    def __init__(self, average):
        self.average = average
        self.stacktrace = ''
        self.cpu_percent = ''
        self.memory = ''
        self.stopped = False

        try:
            thread = Thread(target=self.log_stack_trace)
            thread.start()
        except Exception as e:
            print('Can\'t log traceback information: {}'.format(e))
            traceback.print_exc()

    def stop(self):
        self.stopped = True

    def log_stack_trace(self):
        # average is in ms, sleep requires seconds
        time.sleep(self.average / 1000.0)

        if self.stopped:
            return

            # iterate through every active thread and store the stack-trace
        stack_list = []
        for thread_id, stack in sys._current_frames().items():
            stack_list.append("\n# Thread_id: %s" % thread_id)
            for filename, lineno, name, line in traceback.extract_stack(stack):
                stack_list.append('File: "%s", line %d, in %s' % (filename, lineno, name))
                if line:
                    stack_list.append("  %s" % (line.strip()))

        # Set the values in the object
        self.stacktrace = '<br />'.join(stack_list)
        self.cpu_percent = str(psutil.cpu_percent(interval=None, percpu=True))
        self.memory = str(psutil.virtual_memory())
